/*
    ES Knot
    Copyright 2008 - 2009 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#import "KnotView.h"
#include "knotinterface.h"
#include <sys/time.h>

long curTime() {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec*1000 + tv.tv_usec/1000;
}

void debug(const char *str) {
	NSLog(@"%s", str);
}

@implementation KnotView

- (id)initWithCoder:(NSCoder*)coder {
	knot = NULL;

	contextVersion = 2; // Request an ES2 context if possible
	if ((self = [super initWithCoder:coder])) {
#ifdef USE_OES_2
		if (contextVersion == 2) {
			knot = createKnot2(backingWidth, backingHeight);
		} else
#endif
		{
			knot = createKnot1(backingWidth, backingHeight);
		}
		shading = KnotInterface::UNTEXTURED;

		updating = YES;
		motionBlur = NO;
		radialBlur = 0;
		knot->setUpdateVertices(true);
//		knot->setTexMode(KnotInterface::TEXSPEC);

		UIAccelerometer* theAccelerometer = [UIAccelerometer sharedAccelerometer];
		theAccelerometer.updateInterval = 1 / 50;
		theAccelerometer.delegate = self;
		
		knot->enableRotate(false);
		knot->enableGravity(true);
		
		lastFpsTime = curTime();
		frames = 0;
		[self drawView];
	}
	return self;
}

- (void)drawContent {
	if (!knot) {
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);
		return;
	}
	knot->draw(curTime());
	frames++;
	
	long now = curTime();
	if (now > lastFpsTime + 2000) {
		NSLog(@"%.1f fps", 1000.0f*frames/(now - lastFpsTime));
		frames = 0;
		lastFpsTime = now;
	}
}

- (void)dealloc {
	delete knot;
	[super dealloc];
}

- (void)touchesBegan:(NSSet*)touches withEvent:(UIEvent*)event {
	knot->enableRotate(false);
	knot->enableGravity(false);
	UITouch* touch = [touches anyObject];
	lastPos = [touch locationInView: self];
	distance = 0;
}

- (void)touchesMoved:(NSSet*)touches withEvent:(UIEvent*)event {
	UITouch* touch = [touches anyObject];
	CGPoint pos = [touch locationInView: self];
	int diffX = pos.x - lastPos.x;
	int diffY = pos.y - lastPos.y;
	lastPos = pos;
	knot->rotate(diffX, diffY);
	distance += abs(diffX) + abs(diffY);
}

- (void)touchesEnded:(NSSet*)touches withEvent:(UIEvent*)event {
//	knot->enableRotate(true);
	knot->enableGravity(true);
	if (distance < 5)
		knot->startMorph();
}

- (void)touchesCancelled:(NSSet*)touches withEvent:(UIEvent*)event {
//	knot->enableRotate(true);
	knot->enableGravity(true);
}

- (void)tabBar:(UITabBar *)tabBar didSelectItem:(UITabBarItem *)item {
	NSString* badge = nil;
	switch ([item tag]) {
	case 0:
		shading++;
		if (shading > KnotInterface::CELSHADED)
			shading = KnotInterface::UNTEXTURED;
		knot->setTexMode((KnotInterface::TexMode)shading);
		switch (shading) {
		case KnotInterface::UNTEXTURED:
			badge = @"Color";
			break;
		case KnotInterface::PLAINTEX:
			badge = @"Texture";
			break;
		case KnotInterface::TEXSPEC:
			badge = @"Specular";
			break;
		case KnotInterface::CELSHADED:
			badge = @"Celshaded";
			break;
		}
		break;
	case 1:
		updating = !updating;
		knot->setUpdateVertices(updating);
		badge = updating ? @"On" : @"Off";
		break;
	case 2:
		radialBlur++;
		if (radialBlur > 2)
			radialBlur = 0;
		switch (radialBlur) {
		case 0:
			knot->setRadialBlur(false);
			knot->setRadialGlow(false);
			badge = @"Off";
			break;
		case 1:
			knot->setRadialBlur(true);
			knot->setRadialGlow(false);
			badge = @"On";
			break;
		case 2:
			knot->setRadialBlur(true);
			knot->setRadialGlow(true);
			badge = @"Glow";
			break;
		}
		break;
	case 3:
		motionBlur = !motionBlur;
		knot->setMotionBlur(motionBlur);
		badge = motionBlur ? @"On" : @"Off";
		break;
	}
	[item setBadgeValue: badge];
	[tabBar setSelectedItem: nil];
}

#define kFilteringFactor 0.5

- (void)accelerometer:(UIAccelerometer *)accelerometer didAccelerate:(UIAcceleration *)acceleration {
    accelX = (acceleration.x * kFilteringFactor) + (accelX * (1.0 - kFilteringFactor));
    accelY = (acceleration.y * kFilteringFactor) + (accelY * (1.0 - kFilteringFactor));
    accelZ = (acceleration.z * kFilteringFactor) + (accelZ * (1.0 - kFilteringFactor));
	
	knot->setGravity(accelX, accelY, accelZ);
}

@end
