/*
    ES Knot
    Copyright 2005 - 2008 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include "knot.h"
#include <GLES/gl.h>
#ifndef NO_EGL
#include <GLES/egl.h>
#endif
#ifdef __APPLE__
// This actually maps to OpenGLES/ES1/glext.h
#include <GLES/glext.h>
#define USE_FBO
#endif
#ifdef LINK_EXTS
#define GL_GLEXT_PROTOTYPES
#include <GLES/glext.h>
#endif
#include <math.h>
#include <stdio.h>
#include <string.h>

#ifdef HAS_CUSTOM_GETPROCADDRESS
#define eglGetProcAddress customGetProcAddress
void* customGetProcAddress(const char* name);
#endif

#ifdef CUSTOM_LOAD_OES_1_1
void* customLoad(const char* str);
#endif

void debug(const char *str);

#if !defined(GL_OES_VERSION_1_1) && !defined(GL_VERSION_ES_CM_1_0)
#ifdef USE_OES_1_1
#warning USE_OES_1_1 defined but no ES1.1 header available, disabling
#undef USE_OES_1_1
#endif
#endif

#ifndef GL_MODELVIEW_MATRIX
#define GL_MODELVIEW_MATRIX 0x0BA6
#endif

#define CELSHADE_TEX_MATRIX

#include "knotbase.cpp"


#if defined(USE_OES_1_1) && !defined(NO_EGL)
#define USE_PBUFFER
#endif
#include "pbuffer.cpp"

class RadialBlur {
public:
	RadialBlur(int vpWidth, int vpHeight) {
		pbuffer = new PbufferRTT();
		if (pbuffer->init(256, 256)) {
			textureWidth = pbuffer->getWidth();
			textureHeight = pbuffer->getHeight();
			texture = 0;
		} else {
			delete pbuffer;
			pbuffer = NULL;

			textureWidth = vpWidth;
			textureHeight = vpHeight;
			texture = genEmptyTexture(textureWidth, textureHeight);
		}
	}
	~RadialBlur() {
		if (texture)
			glDeleteTextures(1, &texture);
		delete pbuffer;
	}
	void setLinear(bool linear) {
		if (pbuffer)
			pbuffer->bind();
		else
			glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, linear ? GL_LINEAR : GL_NEAREST);
	}
	void setGlow(bool g) {
		glow = g;
	}

	void init(int& vpWidth, int& vpHeight) {
		origWidth = vpWidth;
		origHeight = vpHeight;

		if (pbuffer) {
			pbuffer->activate();
			vpWidth = textureWidth;
			vpHeight = textureHeight;
		} else {
			if (textureWidth < vpWidth || textureHeight < vpHeight) {
				vpWidth = MIN(textureWidth, vpWidth);
				vpHeight = MIN(textureHeight, vpHeight);
				glViewport(0, 0, vpWidth, vpHeight);
			}
		}

	}
	void finish(int& vpWidth, int& vpHeight) {
		bool cleared = false;

		if (pbuffer) {
			pbuffer->finish();
			glClearColor(0, 0, 0, 0);
			glClear(GL_COLOR_BUFFER_BIT);
			cleared = true;
			pbuffer->bind();
		} else {
			glEnable(GL_TEXTURE_2D);
			glClearColor(F(0), F(0), F(0), F(1));
			glBindTexture(GL_TEXTURE_2D, texture);
//			glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, vpWidth, vpHeight);
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, vpWidth, vpHeight, 0);

			if (vpWidth < origWidth || vpHeight < origHeight) {
				glViewport(0, 0, origWidth, origHeight);
				glClear(GL_COLOR_BUFFER_BIT);
				cleared = true;
			}
		}


#define RADIALLAYERS 5

		TYPE tex[2*4*RADIALLAYERS];
		TYPE verts[2*4*RADIALLAYERS];
		TYPE colors[4*4*RADIALLAYERS];
		GLushort indices[6*RADIALLAYERS];
		static const GLushort rectindices[] = { 0, 1, 2, 0, 2, 3 };

		TYPE texWidth = F(vpWidth)/textureWidth;
		TYPE texHeight = F(vpHeight)/textureHeight;
		for (int i = 0; i < RADIALLAYERS; i++) {
			verts[2*(4*i+0)+0] = F(0);
			verts[2*(4*i+0)+1] = F(0);
			verts[2*(4*i+1)+0] = F(100);
			verts[2*(4*i+1)+1] = F(0);
			verts[2*(4*i+2)+0] = F(100);
			verts[2*(4*i+2)+1] = F(100);
			verts[2*(4*i+3)+0] = F(0);
			verts[2*(4*i+3)+1] = F(100);
			for (int j = 0; j < 6; j++)
				indices[6*i+j] = rectindices[j] + 4*i;
			TYPE alpha = F(1)/(i+1);
			if (glow) {
				if (i == 0)
					alpha = F(1);
				else
					alpha = F(1)/RADIALLAYERS;
			} 
			for (int j = 0; j < 4; j++) {
				colors[4*(4*i+j)+0] = F(1);
				colors[4*(4*i+j)+1] = F(1);
				colors[4*(4*i+j)+2] = F(1);
				colors[4*(4*i+j)+3] = alpha;
			}
			TYPE margin = MULT(F(0.35), F(i))/RADIALLAYERS;
			TYPE xmargin = MULT(texWidth, margin)/2;
			TYPE ymargin = MULT(texHeight, margin)/2;
			tex[2*(4*i+0)+0] = xmargin;
			tex[2*(4*i+0)+1] = ymargin;
			tex[2*(4*i+1)+0] = texWidth - xmargin;
			tex[2*(4*i+1)+1] = ymargin;
			tex[2*(4*i+2)+0] = texWidth - xmargin;
			tex[2*(4*i+2)+1] = texHeight - ymargin;
			tex[2*(4*i+3)+0] = xmargin;
			tex[2*(4*i+3)+1] = texHeight - ymargin;
		}
		glColor4(F(1), F(1), F(1), F(1)/(RADIALLAYERS));


		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(F(0), F(100), F(0), F(100), F(-1), F(1));
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);

		glVertexPointer(2, TYPETOKEN, 0, verts);
		glTexCoordPointer(2, TYPETOKEN, 0, tex);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);


		if (!cleared) {
			if (glow)
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			else
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer(4, TYPETOKEN, 0, colors);
			glDrawElements(GL_TRIANGLES, 6*(RADIALLAYERS-1), GL_UNSIGNED_SHORT, indices+6);
		} else {
			if (glow) {
				glEnableClientState(GL_COLOR_ARRAY);
				glColorPointer(4, TYPETOKEN, 0, colors);
			}
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			glDrawElements(GL_TRIANGLES, 6*RADIALLAYERS, GL_UNSIGNED_SHORT, indices);
		}

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
		glDisableClientState(GL_COLOR_ARRAY);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		vpWidth = origWidth;
		vpHeight = origHeight;
	}
private:
	PbufferRTT* pbuffer;
	GLuint texture;
	int textureWidth, textureHeight;
	int origWidth, origHeight;
	bool glow;
};

class MotionBlur {
public:
	MotionBlur(int vpWidth, int vpHeight) {
		pbuffer1 = new PbufferRTT();
		pbuffer2 = new PbufferRTT();
		if (!pbuffer1->init(256, 256) || !pbuffer2->init(256, 256)) {
			delete pbuffer1;
			delete pbuffer2;
			pbuffer1 = NULL;
			pbuffer2 = NULL;
		}
		if (!pbuffer1) {
			textureWidth = vpWidth;
			textureHeight = vpHeight;
			texture = genEmptyTexture(textureWidth, textureHeight);
		} else
			texture = 0;
	}
	~MotionBlur() {
		glDeleteTextures(1, &texture);
		delete pbuffer1;
		delete pbuffer2;
	}

	void init(int& vpWidth, int& vpHeight) {
		origWidth = vpWidth;
		origHeight = vpHeight;
		if (pbuffer1) {
			pbuffer1->activate();
			textureWidth = vpWidth = pbuffer1->getWidth();
			textureHeight = vpHeight = pbuffer1->getHeight();
		} else {
			if (textureWidth < vpWidth || textureHeight < vpHeight) {
				vpWidth = MIN(textureWidth, vpWidth);
				vpHeight = MIN(textureHeight, vpHeight);
				glViewport(0, 0, vpWidth, vpHeight);
			}
		}
	}
	void finish(int& vpWidth, int& vpHeight, unsigned int diff) {
		TYPE alpha = F(1 - exp(-float(diff)/130));
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(F(0), F(100), F(0), F(100), F(-1), F(1));
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);

		if (pbuffer2)
			pbuffer2->bind();
		else
			glBindTexture(GL_TEXTURE_2D, texture);
		glColor4(F(1), F(1), F(1), F(1) - alpha);

		TYPE tex[2*4];
		TYPE verts[2*4];
		tex[2*0+0] = F(0);
		tex[2*0+1] = F(0);
		tex[2*1+0] = F(vpWidth)/textureWidth;
		tex[2*1+1] = F(0);
		tex[2*2+0] = F(vpWidth)/textureWidth;
		tex[2*2+1] = F(vpHeight)/textureHeight;
		tex[2*3+0] = F(0);
		tex[2*3+1] = F(vpHeight)/textureHeight;
		verts[2*0+0] = F(0);
		verts[2*0+1] = F(0);
		verts[2*1+0] = F(100);
		verts[2*1+1] = F(0);
		verts[2*2+0] = F(100);
		verts[2*2+1] = F(100);
		verts[2*3+0] = F(0);
		verts[2*3+1] = F(100);
		static const GLushort indices[] = { 0, 1, 2, 0, 2, 3 };

		glVertexPointer(2, TYPETOKEN, 0, verts);
		glTexCoordPointer(2, TYPETOKEN, 0, tex);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);

//		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, vpWidth, vpHeight);
		if (pbuffer1) {
			pbuffer1->finish();
			pbuffer1->bind();
			glViewport(0, 0, origWidth, origHeight);
			glDisable(GL_BLEND);
			glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);
		} else {
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, vpWidth, vpHeight, 0);

			if (vpWidth < origWidth || vpHeight < origHeight) {
				glViewport(0, 0, origWidth, origHeight);
				glDisable(GL_BLEND);
				glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);
			}
		}

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		vpWidth = origWidth;
		vpHeight = origHeight;
		
		if (pbuffer1) {
			PbufferRTT* tmp = pbuffer1;
			pbuffer1 = pbuffer2;
			pbuffer2 = tmp;
		}
	}
	void clear() {
		if (pbuffer1 && pbuffer2) {
			pbuffer1->clear();
			pbuffer2->clear();
		} else {
			glBindTexture(GL_TEXTURE_2D, texture);
			GLubyte *data = new GLubyte[textureWidth*textureHeight*3];
			memset(data, 0, sizeof(GLubyte)*textureWidth*textureHeight*3);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, textureWidth, textureHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
			delete [] data;
		}
	}
private:
	GLuint texture;
	int textureWidth, textureHeight;
	int origWidth, origHeight;
	PbufferRTT* pbuffer1, *pbuffer2;
};

Knot::Knot(int w, int h) {
	width = w;
	height = h;
	updateBumps = false;
	doMotionBlur = false;
	doRadialBlur = false;
	morphing = false;
	rotating = true;
	gravitySet = false;

	texture = 0;
	celTexture = 0;

	motionBlur = NULL;
	radialBlur = NULL;

	ext_glQueryMatrixxOES = NULL;
	if (hasExtension("OES_query_matrix") || hasExtension("GL_OES_query_matrix")) {
#ifndef NO_EGL
		ext_glQueryMatrixxOES = (GLbitfield (*)(GLfixed[], GLint[])) eglGetProcAddress("glQueryMatrixxOES");
#elif defined(LINK_EXTS)
		ext_glQueryMatrixxOES = glQueryMatrixxOES;
#endif
	}
	ext_glGetTypev = NULL;
#if !defined(USE_OES_1_1) && defined(LINK_OES_1_1)
	if (hasVersion(1, 1))
		ext_glGetTypev = glGetTypev;
#elif !defined(USE_OES_1_1) && defined(CUSTOM_LOAD_OES_1_1)
	if (hasVersion(1, 1))
		ext_glGetTypev = (void (*)(GLenum, TYPE*)) customLoad(GLGETTYPEV_NAME);
#endif

	precalc();

	reinit(UNTEXTURED);
}

bool Knot::getMatrixSupported() {
#ifndef USE_OES_1_1
	if (!ext_glQueryMatrixxOES && !ext_glGetTypev)
		return false;
#endif
	return true;
}

void Knot::setUpdateVertices(bool update) {
	updateBumps = update;
	setTexMode(texMode);
	if (!updateBumps)
		calcDynVerticesNormals(vertices, normals, ringcoords);
}

void Knot::setMotionBlur(bool blur) {
	doMotionBlur = blur;
	setDrawState();
	if (blur)
		motionBlur->clear();
}

void Knot::setRadialBlur(bool blur) {
	doRadialBlur = blur;
	setDrawState();
}

void Knot::setRadialGlow(bool g) {
	radialBlur->setGlow(g);
}

void Knot::setSize(int w, int h) {
	width = w;
	height = h;
	initProjection();
}

void Knot::reinit() {
	reinit(texMode);
}

void Knot::enableRotate(bool r) {
	rotating = r;
}

void Knot::enableGravity(bool g) {
	gravityEnabled = g;
	gravitySet = false;
}

void Knot::rotate(int x, int y) {
	glMatrixMode(GL_MODELVIEW);
	TYPE modelview[16];
	if (!getModelview(modelview))
		return;
	glLoadIdentity();
	glTranslate(F(0), F(0), F(-10));
	modelview[12] = modelview[13] = modelview[14] = F(0);

	int minDim = width;
	if (height < minDim)
		minDim = height;
	glRotate(F(180)*x/minDim, F(0), F(1), F(0));
	glRotate(F(180)*y/minDim, F(1), F(0), F(0));
	glMultMatrix(modelview);
}

void Knot::clearContext() {
	glBindTexture(GL_TEXTURE_2D, 0);

	glDeleteTextures(1, &texture);
	glDeleteTextures(1, &celTexture);
	delete motionBlur;
	delete radialBlur;
	texture = 0;
	celTexture = 0;
	motionBlur = NULL;
	radialBlur = NULL;
}

void Knot::reinit(TexMode mode) {
	glViewport(0, 0, width, height);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslate(F(0), F(0), F(-10));

	initProjection();

	genTexture();
	genCelTexture();

	radialBlur = new RadialBlur(width, height);
	motionBlur = new MotionBlur(width, height);
//	radialBlur->setLinear(true);

	lasttime = (unsigned int)-1;
	texOffsetX = texOffsetY = 0;
	bumpOffset = 0;

	setTexMode(mode);
}

void Knot::initProjection() {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float zn = 5;
	float planeSize = zn*tan((54/2)*M_PI/180);
	float fw, fh;
	if (width < height) {
		fw = planeSize;
		fh = fw*height/width;
	} else {
		fh = planeSize;
		fw = fh*width/height;
	}
	glFrustum(F(-fw), F(fw), F(-fh), F(fh), F(zn), F(15));
}

static const TYPE lightPos[] = { F(-3), F(10), F(0), F(1) };
static const TYPE lightAmbient[] = { F(0.2), F(0.2), F(0.2), F(1) };
static const TYPE lightDiffuse[] = { F(1), F(1), F(1), F(1) };

void Knot::setTexMode(TexMode mode) {
	if (mode == CELSHADED) {
		if (!getMatrixSupported())
			return;
	}

	glShadeModel(GL_SMOOTH);
	glClearDepth(F(1));
	glDepthRange(F(0), F(1));
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	static const TYPE lightSpecular[] = { F(1), F(1), F(1), F(1) };
	static const TYPE matSpecular[] = { F(0.8), F(0.8), F(0.8), F(1) };

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glLightv(GL_LIGHT0, GL_POSITION, lightPos);
	glPopMatrix();

	// if color material happens to be enabled, disable it
	// to make sure the material parameters actually are updated
	glDisable(GL_COLOR_MATERIAL);

	glEnable(GL_LIGHT0);
	glLightv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
	glLightv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
	glLightv(GL_LIGHT0, GL_SPECULAR, lightSpecular);
	glMaterialv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecular);
	glMaterial(GL_FRONT_AND_BACK, GL_SHININESS, F(35));

	static const TYPE ambient[] = { F(0.2), F(0.2), F(0.2), F(1) };
	static const TYPE diffuse[] = { F(0.8), F(0.8), F(0.8), F(1) };

	// disable the scene global ambient and enable ambient
	// on the light source, since n93 doesn't seem to use
	// the material ambient color with the scene global ambient
	static const TYPE sceneAmbient[] = { F(0), F(0), F(0), F(1) };
	glLightModelv(GL_LIGHT_MODEL_AMBIENT, sceneAmbient);

	glMaterialv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
	glMaterialv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);

	texMode = mode;
	setDrawState();
}

void Knot::setDrawState() {
	glEnable(GL_LIGHTING);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_COLOR_MATERIAL);
#ifdef SCREENSHOT
	glClearColor(F(0), F(0), F(0), F(0));
#else
	glClearColor(F(0), F(0), F(0), F(1));
#endif
	glColor4(F(1), F(1), F(1), F(1));
	glEnable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, texture);

	glVertexPointer(3, TYPETOKEN, 0, vertices);
	glNormalPointer(TYPETOKEN, 0, normals);
	glTexCoordPointer(2, TYPETOKEN, 0, texcoords);

	switch (texMode) {
	case UNTEXTURED:
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnable(GL_COLOR_MATERIAL);
		glColor4(F(0.1), F(0.1), F(1), F(1));
		break;

	case PLAINTEX:
	case TEXSPEC:
		glEnable(GL_TEXTURE_2D);
		break;

	case CELSHADED:
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnable(GL_TEXTURE_2D);
#ifdef SCREENSHOT
		glClearColor(F(1), F(1), F(1), F(0));
#else
		glClearColor(F(1), F(1), F(1), F(1));
#endif
		glDisable(GL_LIGHTING);
		glLineWidth(F(3));
		glBindTexture(GL_TEXTURE_2D, celTexture);
#ifdef CELSHADE_TEX_MATRIX
		glTexCoordPointer(3, TYPETOKEN, 0, normals);
#else
		glTexCoordPointer(2, TYPETOKEN, 0, celcoords);
#endif
		break;
	case MAX_TEXMODE:
		break;
	}
}

Knot::~Knot() {
	glBindTexture(GL_TEXTURE_2D, 0);
	glDeleteTextures(1, &texture);
	glDeleteTextures(1, &celTexture);
	delete motionBlur;
	delete radialBlur;
}

void Knot::startMorph() {
	if (morphing) return;
	morphing = true;
	morphStart = lasttime;
}

void Knot::draw(unsigned int msec) {
	if (lasttime == (unsigned int)-1)
		lasttime = msec;
	unsigned int diff = msec - lasttime;
	lasttime = msec;

	TYPE *verts = vertices;
	TYPE *norms = normals;

	if (morphing) {
#define MORPHLENGTH 1200
		if (msec - morphStart > MORPHLENGTH) {
			morphing = false;
			if (ringcoords == ringcoords23)
				ringcoords = ringcoords25;
			else
				ringcoords = ringcoords23;
			calcDynVerticesNormals(vertices, normals, ringcoords);
		} else {
			TYPE *next = ringcoords23;
			if (ringcoords == ringcoords23)
				next = ringcoords25;
			morphMatrices(ringcoordstemp, ringcoords, next, F(msec-morphStart)/MORPHLENGTH);
			calcDynVerticesNormals(verts, normals, ringcoordstemp);
		}
	} else if (updateBumps)
		calcDynVertices();

	int vpwidth = width;
	int vpheight = height;
	if (doRadialBlur || doMotionBlur)
		setDrawState();

	if (doMotionBlur)
		motionBlur->init(vpwidth, vpheight);

	if (doRadialBlur)
		radialBlur->init(vpwidth, vpheight);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (rotating) {
		glMatrixMode(GL_MODELVIEW);
		glRotate(F(90)*diff/1000, F(sin(msec*M_PI/3000)), F(sin(msec*M_PI/600)), F(cos(msec*M_PI/900)));
	} else if (gravityEnabled && gravitySet) {
		AVG3(curGravity, curGravity, targetGravity);
		NORMALIZE3(curGravity);
		rotateGravity();
	}

	texOffsetY += F(diff)/30000;
	bumpOffset += F(diff)/4000;
#ifdef USING_FIXED
	texOffsetY &= 0x0000FFFF;
	bumpOffset &= 0x0000FFFF;
#else
	if (texOffsetY > 1) texOffsetY -= int(texOffsetY);
	if (bumpOffset > 1) bumpOffset -= int(bumpOffset);
#endif

	if (texMode == CELSHADED) {
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();

		updateOutlineVertices(outlineVertices, verts, norms);

		glCullFace(GL_FRONT);
		glColor4(F(0), F(0), F(0), F(1));
		glVertexPointer(3, TYPETOKEN, 0, outlineVertices);
		glDisable(GL_TEXTURE_2D);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

#ifdef CELSHADE_TEX_MATRIX
		TYPE light[4];
		getLightInverse(light);
		NORMALIZE3(light);
		TYPE *dir = light;

		TYPE texMatrix[16];
		memset(texMatrix, 0, sizeof(texMatrix));
		texMatrix[0] = -dir[0];
		texMatrix[4] = -dir[1];
		texMatrix[8] = -dir[2];
		texMatrix[12] = F(1);
		texMatrix[15] = F(1);
		glMatrixMode(GL_TEXTURE);
		glLoadMatrix(texMatrix);
#else
		updateCelCoords(verts, norms);
#endif

		glCullFace(GL_BACK);
		glColor4(F(1), F(1), F(1), F(1));
		glVertexPointer(3, TYPETOKEN, 0, vertices);
		glEnable(GL_TEXTURE_2D);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);


	} else {

		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glTranslate(F(0), MULTY*texOffsetY, F(0));

		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

		if (texMode == TEXSPEC) {
			static const TYPE zero[] = { F(0), F(0), F(0), F(0) };

			glEnable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glBlendFunc(GL_ONE, GL_ONE);
			glLightv(GL_LIGHT0, GL_DIFFUSE, zero);
			glLightv(GL_LIGHT0, GL_AMBIENT, zero);

			glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

			glDisable(GL_BLEND);
			glEnable(GL_TEXTURE_2D);
			glLightv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
			glLightv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
		}
	}


	if (doRadialBlur)
		radialBlur->finish(vpwidth, vpheight);

	if (doMotionBlur)
		motionBlur->finish(vpwidth, vpheight, diff);
}

bool Knot::getModelview(TYPE *matrix) {
#ifdef USE_OES_1_1
	glGetTypev(GL_MODELVIEW_MATRIX, matrix);
#else
	if (ext_glGetTypev) {
		ext_glGetTypev(GL_MODELVIEW_MATRIX, matrix);
		return true;
	}
	if (!ext_glQueryMatrixxOES)
		return false;
	GLfixed mantissa[16];
	GLint exponent[16];
	glMatrixMode(GL_MODELVIEW);
	ext_glQueryMatrixxOES(mantissa, exponent);
	for (int i = 0; i < 16; i++) {
		if (exponent[i] >= 0)
			matrix[i] = FIXED_TO_NATIVE(mantissa[i] << exponent[i]);
		else
			matrix[i] = FIXED_TO_NATIVE(mantissa[i] >> (-exponent[i]));
	}
#endif
	return true;
}

void Knot::getLightInverse(TYPE* light) {
	TYPE rot[16], trans[16], lightMatrix[16];
	TYPE modelview[16];
	memset(rot, 0, sizeof(rot));
	memset(trans, 0, sizeof(rot));
	memset(lightMatrix, 0, sizeof(rot));

	glMatrixMode(GL_MODELVIEW);
	getModelview(modelview);
	glPushMatrix();
	glLoadIdentity();

	// make rot and trans identity matrices
	for (int i = 0; i < 4; i++)
		rot[i+i*4] = trans[i+i*4] = F(1);

	// rot is the rotational part from modelview, transposed,
	// which also is the inverse of the rotational part
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			rot[i+j*4] = modelview[j+i*4];
	for (int i = 0; i < 3; i++)
		trans[i+3*4] = -modelview[i+3*4];
	for (int i = 0; i < 4; i++)
		lightMatrix[i] = lightPos[i];

	glMultMatrix(rot);
	glMultMatrix(trans);
	glMultMatrix(lightMatrix);

	getModelview(lightMatrix);
	glPopMatrix();

	for (int i = 0; i < 4; i++)
		light[i] = lightMatrix[i];
}

void Knot::setGravity(float x, float y, float z) {
	targetGravity[0] = F(x);
	targetGravity[1] = F(y);
	targetGravity[2] = F(z);

	if (!gravitySet) {
		COPY3(curGravity, targetGravity);
		complementVect[0] = F(1);
		complementVect[1] = F(0);
		complementVect[2] = F(0);
		extraRotInited = false;
	}
	gravitySet = true;
}

void Knot::rotateGravity() {
	TYPE thirdVect[3];

	cross3(thirdVect, curGravity, complementVect);
	NORMALIZE3(thirdVect);
	cross3(complementVect, thirdVect, curGravity);
	NORMALIZE3(complementVect);
	
	TYPE matrix[] = {
		complementVect[0], complementVect[1], complementVect[2], F(0),
		thirdVect[0], thirdVect[1], thirdVect[2], F(0),
		curGravity[0], curGravity[1], curGravity[2], F(0),
		F(0), F(0), F(-10), F(1)
	};
	
	glMatrixMode(GL_MODELVIEW);

	if (!extraRotInited) {
		TYPE curRot[16];
		getModelview(curRot);
		for (int i = 0; i < 3; i++)
			curRot[12+i] = F(0);
		
		TYPE invStartPos[16];
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++)
				invStartPos[i+4*j] = matrix[j+4*i];
		invStartPos[3] = invStartPos[7] = invStartPos[11] = F(0);
		invStartPos[12] = invStartPos[13] = invStartPos[14] = F(0);
		invStartPos[15] = F(1);
		
		glPushMatrix();
		glLoadMatrix(invStartPos);
		glMultMatrix(curRot);
		getModelview(extraRotMatrix);
		glPopMatrix();
		extraRotInited = true;
	}
	
	glLoadMatrix(matrix);
	glMultMatrix(extraRotMatrix);
}

KnotInterface* createKnot1(int w, int h) {
	return new Knot(w, h);
}

